# Copyright (c) 2025 Thomas Goirand <zigo@debian.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import threading
from oslo_db.sqlalchemy import enginefacade
from oslo_config import cfg

# Global variables for engine facade
_LOCK = threading.Lock()
_MAIN_CONTEXT_MANAGER = None

def _get_main_context_manager():
    global _LOCK, _MAIN_CONTEXT_MANAGER

    if _MAIN_CONTEXT_MANAGER is not None:
        return _MAIN_CONTEXT_MANAGER

    with _LOCK:
        if _MAIN_CONTEXT_MANAGER is None:
            _MAIN_CONTEXT_MANAGER = enginefacade.transaction_context()
    return _MAIN_CONTEXT_MANAGER

def init_database_config(CONF):
    """Initialize database configuration with the given CONF object."""
    context_manager = _get_main_context_manager()
    context_manager.configure(
        connection=CONF.database.connection
    )

def get_session():
    """Get database session using oslo.db enginefacade."""
    return _get_main_context_manager().writer.get_sessionmaker()()

def get_engine():
    """Get database engine using oslo.db enginefacade."""
    return _get_main_context_manager().writer.get_engine()

def get_reader_session():
    """Get read-only database session."""
    return _get_main_context_manager().reader.get_sessionmaker()()

def get_writer_session():
    """Get write database session."""
    return _get_main_context_manager().writer.get_sessionmaker()()



from unittest import mock
import json
import uuid

from vmms.tests import base


class TestVMMSServiceAPI(base.VMMSTestCase):
    """Test VMMS service API endpoints."""

    def setUp(self):
        super(TestVMMSServiceAPI, self).setUp()
        
        # Mock database session
        self.session_patcher = mock.patch('vmms.db.sqlalchemy.database.get_session')
        self.mock_session = self.session_patcher.start()
        self.addCleanup(self.session_patcher.stop)
        
        # Mock the config module to avoid initialization
        self.config_module_patcher = mock.patch('vmms.config.init_config')
        self.mock_init_config = self.config_module_patcher.start()
        self.mock_init_config.return_value = mock.Mock()  # Return mock config
        self.addCleanup(self.config_module_patcher.stop)
        
        # Mock get_vm_details_from_source_cloud
        self.vm_details_patcher = mock.patch('vmms.api.get_vm_details_from_source_cloud')
        self.mock_get_vm_details = self.vm_details_patcher.start()
        self.addCleanup(self.vm_details_patcher.stop)
        
        # Mock policy enforcement to always allow (for functional tests)
        self.policy_patcher = mock.patch('vmms.policy.enforcer.enforce_policy')
        self.mock_enforce_policy = self.policy_patcher.start()
        self.mock_enforce_policy.return_value = True  # Always allow policy
        self.addCleanup(self.policy_patcher.stop)

    def test_healthcheck_endpoint(self):
        """Test health check endpoint."""
        response = self.client.get('/healthcheck')
        self.assertEqual(response.status_code, 200)
        data = json.loads(response.data)
        self.assertEqual(data['status'], 'ok')
        self.assertEqual(data['service'], 'vmms')

    def test_add_vm_missing_vm_identifier(self):
        """Test adding VM without vm_identifier."""
        headers = self._create_test_headers(roles=['vmms_operator'])
        response = self.client.post('/v2/vms', headers=headers, json={})
        self.assertEqual(response.status_code, 400)
        data = json.loads(response.data)
        self.assertIn('error', data)

    def test_invalid_migration_id_format(self):
        """Test API with invalid migration ID format."""
        # Mock the is_valid_uuid function to return False for invalid IDs
        with mock.patch('vmms.api.is_valid_uuid', return_value=False):
            headers = self._create_test_headers(roles=['vmms_operator'])
            response = self.client.get('/v2/vms/invalid-id', headers=headers)
            self.assertEqual(response.status_code, 400)


class TestVMMSPolicyDeny(base.VMMSTestCase):
    """Test that policy denials work correctly."""

    def setUp(self):
        super(TestVMMSPolicyDeny, self).setUp()
        
        # Mock database session
        self.session_patcher = mock.patch('vmms.db.sqlalchemy.database.get_session')
        self.mock_session = self.session_patcher.start()
        self.addCleanup(self.session_patcher.stop)
        
        # Mock the config module to avoid initialization
        self.config_module_patcher = mock.patch('vmms.config.init_config')
        self.mock_init_config = self.config_module_patcher.start()
        self.mock_init_config.return_value = mock.Mock()  # Return mock config
        self.addCleanup(self.config_module_patcher.stop)
        
        # Mock get_vm_details_from_source_cloud
        self.vm_details_patcher = mock.patch('vmms.api.get_vm_details_from_source_cloud')
        self.mock_get_vm_details = self.vm_details_patcher.start()
        self.addCleanup(self.vm_details_patcher.stop)
        
        # Mock the policy enforcer to raise exceptions for denied access
        self.policy_patcher = mock.patch('vmms.policy.enforcer.enforce_policy')
        self.mock_enforce_policy = self.policy_patcher.start()
        self.addCleanup(self.policy_patcher.stop)

    def test_viewer_cannot_add_vm(self):
        """Test that vmms_viewer users cannot add VMs."""
        # Make policy enforcement raise an exception for add operations by viewers
        def mock_enforce_policy_side_effect(*args, **kwargs):
            # Simulate a policy denial by raising an exception
            from oslo_policy import policy as oslo_policy
            raise oslo_policy.PolicyNotAuthorized('vmms:add', {}, {})
        
        self.mock_enforce_policy.side_effect = mock_enforce_policy_side_effect
        
        headers = self._create_test_headers(roles=['vmms_viewer'])
        response = self.client.post('/v2/vms', headers=headers, json={
            'vm_identifier': 'test-vm'
        })
        # Should be forbidden (the decorator catches the exception and returns 403)
        self.assertEqual(response.status_code, 403)



from unittest import mock

from vmms.tests import base
from vmms.policy import enforcer


class TestPolicyEnforcer(base.VMMSPolicyTestCase):
    """Test policy rule enforcement."""

    def test_vmms_viewer_can_list(self):
        """Test that vmms_viewer can list VMs."""
        creds = self._create_creds(roles=['vmms_viewer'])
        target = self._create_target()
        
        self.assertRuleIsEnforced('vmms:list', creds, target, should_pass=True)

    def test_vmms_viewer_cannot_set(self):
        """Test that vmms_viewer cannot set VMs."""
        creds = self._create_creds(roles=['vmms_viewer'])
        target = self._create_target()
        
        self.assertRuleIsEnforced('vmms:update', creds, target, should_pass=False)

    def test_vmms_operator_can_list(self):
        """Test that vmms_operator can list VMs."""
        creds = self._create_creds(roles=['vmms_operator'])
        target = self._create_target()
        
        self.assertRuleIsEnforced('vmms:list', creds, target, should_pass=True)

    def test_vmms_operator_can_set(self):
        """Test that vmms_operator can set VMs."""
        creds = self._create_creds(roles=['vmms_operator'])
        target = self._create_target()
        
        self.assertRuleIsEnforced('vmms:update', creds, target, should_pass=True)

    def test_admin_can_list(self):
        """Test that admin can list VMs."""
        creds = self._create_creds(roles=['admin'])
        target = self._create_target()
        
        self.assertRuleIsEnforced('vmms:list', creds, target, should_pass=True)

    def test_admin_can_set(self):
        """Test that admin can set VMs."""
        creds = self._create_creds(roles=['admin'])
        target = self._create_target()
        
        self.assertRuleIsEnforced('vmms:update', creds, target, should_pass=True)

    def test_regular_user_cannot_list(self):
        """Test that regular user cannot list VMs."""
        creds = self._create_creds(roles=['member'])
        target = self._create_target()
        
        self.assertRuleIsEnforced('vmms:list', creds, target, should_pass=False)

    def test_regular_user_cannot_set(self):
        """Test that regular user cannot set VMs."""
        creds = self._create_creds(roles=['member'])
        target = self._create_target()
        
        self.assertRuleIsEnforced('vmms:update', creds, target, should_pass=False)


class TestPolicyDecorator(base.VMMSPolicyTestCase):
    """Test policy decorator functionality."""

    def test_require_policy_factory_creates_decorator(self):
        """Test that require_policy_factory creates a decorator."""
        def mock_get_config():
            # Return a simple mock config for testing
            mock_conf = mock.Mock()
            return mock_conf
            
        decorator_factory = enforcer.require_policy_factory(mock_get_config)
        decorator = decorator_factory('vmms:list')
        
        self.assertIsNotNone(decorator)



from logging.config import fileConfig
from sqlalchemy import engine_from_config
from sqlalchemy import pool
from alembic import context
import sys
import os

# Add project root to path
current_dir = os.path.dirname(os.path.abspath(__file__))
project_root = os.path.join(current_dir, '..', '..', '..', '..')
sys.path.insert(0, project_root)

from vmms.db.sqlalchemy.models import Base

# Alembic Config object
config = context.config

# Interpret the config file for Python logging
if config.config_file_name is not None:
    fileConfig(config.config_file_name)

target_metadata = Base.metadata

def run_migrations_offline():
    """Run migrations in 'offline' mode."""
    url = config.get_main_option("sqlalchemy.url")
    context.configure(
        url=url,
        target_metadata=target_metadata,
        literal_binds=True,
        dialect_opts={"paramstyle": "named"},
    )

    with context.begin_transaction():
        context.run_migrations()

def run_migrations_online():
    """Run migrations in 'online' mode."""
    connectable = engine_from_config(
        config.get_section(config.config_ini_section),
        prefix="sqlalchemy.",
        poolclass=pool.NullPool,
    )

    with connectable.connect() as connection:
        context.configure(
            connection=connection, target_metadata=target_metadata
        )

        with context.begin_transaction():
            context.run_migrations()

if context.is_offline_mode():
    run_migrations_offline()
else:
    run_migrations_online()



# Copyright (c) 2025 Thomas Goirand <zigo@debian.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
import os
import signal
import socket
import sys
import threading
import time

from oslo_config import cfg
from oslo_log import log as logging
import tooz
from tooz import coordination

from vmms import config
from vmms.db.sqlalchemy.database import get_session as get_db_session
from vmms.db.sqlalchemy.database import init_database_config
from vmms.db.sqlalchemy.models import VMMigration, MigrationState
from vmms.mistral_client import MistralClient

# Initialize logging
LOG = logging.getLogger(__name__)

def sd_notify(state: str):
    """Send a notification to systemd."""
    notify_socket = os.getenv("NOTIFY_SOCKET")
    if not notify_socket:
        return False

    # Abstract namespace socket if starts with '@'
    if notify_socket[0] == '@':
        notify_socket = '\0' + notify_socket[1:]

    sock = socket.socket(socket.AF_UNIX, socket.SOCK_DGRAM)
    try:
        sock.connect(notify_socket)
        sock.sendall(state.encode())
    except Exception as e:
        LOG.debug(f"🐛 Failed to notify systemd: {e}")
        return False
    finally:
        sock.close()
    return True


class WorkerService(object):
    """Main worker service for VM migration scheduling."""

    def __init__(self, CONF):
        self.CONF = CONF
        self.node_id = CONF.worker.my_hostname.encode('utf-8')
        self.group_name = b'vm_migration_workers'
        self.running = False
        self.coordinator = None
        self.scheduler_lock = None
        self._stop_event = threading.Event()

        # Initialize components
        self.mistral_client = MistralClient(CONF)

        # Setup signal handlers
        signal.signal(signal.SIGTERM, self._signal_handler)
        signal.signal(signal.SIGINT, self._signal_handler)

    def _signal_handler(self, signum, frame):
        """Handle shutdown signals."""
        LOG.info(f"⚙ VMMS worker is shutting down...")
        self.stop()

    def initialize_coordination(self):
        """Initialize tooz coordination."""
        if not self.CONF.worker.coordination_url:
            LOG.error(f"⧱ Coordination URL not configured")
            return False

        try:
            self.coordinator = coordination.get_coordinator(
                self.CONF.worker.coordination_url,
                self.node_id
            )
            self.coordinator.start()

            # Join the worker group
            try:
                self.coordinator.join_group(self.group_name).get()
            except coordination.GroupNotCreated:
                # Try to create the group (first worker)
                try:
                    self.coordinator.create_group(self.group_name).get()
                    self.coordinator.join_group(self.group_name).get()
                except coordination.GroupAlreadyExist:
                    # Another worker created it, join normally
                    self.coordinator.join_group(self.group_name).get()
            except coordination.ToozError as e:
                LOG.error(f"⧱ Failed to join coordination group: {e}")
                return False

            # Create scheduler lock
            self.scheduler_lock = self.coordinator.get_lock(b'vm_scheduler_lock')

            LOG.info(f"⚙ Successfully initialized coordination with node_id: {self.CONF.worker.my_hostname}")
            return True

        except coordination.ToozError as e:
            LOG.error(f"⧱ Tooz coordination error: {e}")
            return False
        except Exception as e:
            LOG.error(f"⧱ Failed to initialize coordination: {e}")
            return False

    def start(self):
        """Start the worker service."""
        LOG.info(f"⚙ VMMS worker is starting up...")
        LOG.info(f"⚙ VMMS worker Service (node_id: {self.CONF.worker.my_hostname})")

        # Initialize database configuration
        init_database_config(self.CONF)

        if not self.initialize_coordination():
            LOG.error(f"⧱ Failed to initialize coordination, exiting")
            return False

        # Check if Mistral client initialized successfully
        if not self.mistral_client.client:
            LOG.error(f"⧱ Mistral client failed to initialize, exiting")
            return False

        self.running = True

        # Notify systemd that we're ready
        sd_notify("READY=1")
        LOG.info(f"⚙ VMMS worker is ready.")

        # Main worker loop
        while self.running and not self._stop_event.is_set():
            try:
                self.run_periodic_check()
                self._stop_event.wait(self.CONF.worker.check_interval)
            except Exception as e:
                LOG.error(f"⧱ Error in periodic check: {e}", exc_info=True)
                self._stop_event.wait(10)  # Wait before retry

        self.cleanup()
        return True

    def stop(self):
        """Stop the worker service."""
        LOG.info(f"⚙ VMMS worker is shutting down...")
        self.running = False
        self._stop_event.set()
        sd_notify("STOPPING=1")

    def cleanup(self):
        """Cleanup resources."""
        if self.coordinator:
            try:
                self.coordinator.leave_group(self.group_name).get()
                self.coordinator.stop()
            except Exception as e:
                LOG.warning(f"⚠ Error during coordinator cleanup: {e}")

    def parse_migration_result(self, execution_output):
        """Parse migration result from execution output.
        
        Looks for 'Migration successful.' or 'Migration failed.' in the output.
        Returns True for success, False for failure or undetermined result.
        """
        LOG.debug(f"🐛 Execution output type: {type(execution_output)}, content: {execution_output}")
        
        if not execution_output:
            LOG.error(f"⧱ Cannot parse migration result: No execution output provided")
            return False
            
        try:
            # Get the output from execution - check if it's already the output dict
            if isinstance(execution_output, dict):
                if 'output' in execution_output:
                    workflow_output = execution_output['output']
                    LOG.debug(f"🐛 Workflow output type: {type(workflow_output)}, content: {workflow_output}")
                else:
                    # Might be the output directly
                    workflow_output = execution_output
                    LOG.debug(f"🐛 Using execution_output as workflow_output directly")
            else:
                LOG.error(f"⧱ Execution output is not a dict: {type(execution_output)}")
                return False
            
            # Parse JSON string if workflow_output is a string
            if isinstance(workflow_output, str):
                import json
                try:
                    workflow_output = json.loads(workflow_output)
                    LOG.debug(f"🐛 Parsed JSON workflow output type: {type(workflow_output)}, content: {workflow_output}")
                except json.JSONDecodeError as e:
                    LOG.error(f"⧱ Failed to parse workflow output JSON: {e}")
                    return False
            elif not isinstance(workflow_output, dict):
                LOG.error(f"⧱ Workflow output is not a dict or JSON string: {type(workflow_output)}")
                return False
            
            # Get the SSH result
            if isinstance(workflow_output, dict):
                if 'stdout' in workflow_output:
                    # Direct stdout access (simplified case)
                    stdout = workflow_output['stdout']
                    LOG.debug(f"🐛 Stdout from workflow_output: '{stdout}'")
                elif 'result' in workflow_output:
                    ssh_result = workflow_output['result']
                    LOG.debug(f"🐛 SSH result type: {type(ssh_result)}, content: {ssh_result}")
                    
                    # Handle different result formats
                    if isinstance(ssh_result, dict):
                        stdout = ssh_result.get('stdout', '')
                        LOG.debug(f"🐛 Stdout from dict: '{stdout}'")
                    elif isinstance(ssh_result, str):
                        stdout = ssh_result
                        LOG.debug(f"🐛 Stdout as string: '{stdout}'")
                    else:
                        stdout = str(ssh_result)
                        LOG.debug(f"🐛 Stdout converted to string: '{stdout}'")
                else:
                    LOG.error(f"⧱ No stdout or result in workflow output")
                    return False
            else:
                LOG.error(f"⧱ Unexpected workflow output type after JSON parsing: {type(workflow_output)}")
                return False
            
            # Parse the output for migration result messages
            if stdout:
                lines = stdout.strip().split('\n')
                LOG.debug(f"🐛 Parsed lines: {lines}")
                
                # Check lines from bottom up for the result messages
                for line in reversed(lines):
                    line_stripped = line.strip()
                    if 'Migration successful.' in line_stripped:
                        LOG.debug(f"🐛 Migration script reported success in line: '{line_stripped}'")
                        return True
                    elif 'Migration failed.' in line_stripped:
                        LOG.debug(f"🐛 Migration script reported failure in line: '{line_stripped}'")
                        return False
                
                # No clear success/failure message found
                LOG.error(f"⧱ Migration script output found but no clear success/failure message. Last line: '{lines[-1].strip() if lines else 'NONE'}'")
                return False
            else:
                LOG.error(f"⧱ No stdout found in SSH result")
                return False
                
        except Exception as e:
            LOG.error(f"⧱ Error parsing migration result: {e}", exc_info=True)
            return False

    def run_periodic_check(self):
        """Run the periodic check cycle."""
        LOG.debug(f"🐛 Running periodic check cycle")

        # Always monitor running migrations
        try:
            self.monitor_running_migrations()
        except Exception as e:
            LOG.error(f"⧱ Error monitoring running migrations: {e}", exc_info=True)

        # Attempt to become scheduler if we can acquire the lock
        if self.attempt_scheduler_role():
            try:
                self.schedule_new_migrations()
            except Exception as e:
                LOG.error(f"⧱ Error scheduling new migrations: {e}", exc_info=True)
            finally:
                self.release_scheduler_role()

    def attempt_scheduler_role(self):
        """Attempt to acquire scheduler role."""
        if not self.scheduler_lock:
            return False

        try:
            return self.scheduler_lock.acquire(blocking=False)
        except Exception as e:
            LOG.warning(f"⚠ Failed to acquire scheduler lock: {e}")
            return False

    def release_scheduler_role(self):
        """Release scheduler role."""
        if self.scheduler_lock and self.scheduler_lock.acquired:
            try:
                self.scheduler_lock.release()
            except Exception as e:
                LOG.warning(f"⚠ Failed to release scheduler lock: {e}")

    def monitor_running_migrations(self):
        """Monitor MIGRATING migrations and update their status."""
        LOG.debug(f"🐛 Monitoring running migrations")

        session = get_db_session()
        try:
            # Find all migrations in MIGRATING state that we own
            migrating_migrations = session.query(VMMigration).filter(
                VMMigration.state == MigrationState.MIGRATING,
                VMMigration.workflow_exec.isnot(None)
            ).all()
            
            for migration in migrating_migrations:
                try:
                    self.check_and_update_migration_status(session, migration)
                except Exception as e:
                    LOG.error(f"⧱ Error checking migration {migration.id}: {e}")

        finally:
            pass  # No need to close session with oslo.db

    def check_and_update_migration_status(self, session, migration):
        """Check workflow execution status and update migration state."""
        if not migration.workflow_exec:
            LOG.debug(f"🐛 Migration {migration.id} has no workflow execution, skipping")
            return

        try:
            LOG.debug(f"🐛 Checking workflow execution {migration.workflow_exec} for migration {migration.id}")

            execution = self.mistral_client.get_execution(migration.workflow_exec)
            if not execution:
                LOG.warning(f"⚠ Could not retrieve execution {migration.workflow_exec} for migration {migration.id}")
                return

            execution_state = execution.get('state', '').upper()
            LOG.debug(f"🐛 Migration {migration.id} workflow execution {migration.workflow_exec} is in state: {execution_state}")

            if execution_state in ['SUCCESS', 'DONE']:
                # Get execution output to check actual migration result
                execution_output = self.mistral_client.get_execution_output(migration.workflow_exec)
                
                if self.parse_migration_result(execution_output):
                    LOG.info(f"⚙ Migration {migration.id} completed successfully (execution: {migration.workflow_exec})")
                    migration.state = MigrationState.MIGRATED
                else:
                    LOG.info(f"⚙ Migration {migration.id} command failed (execution: {migration.workflow_exec})")
                    migration.state = MigrationState.ERROR
                    
                session.commit()
                LOG.info(f"⚙ Migration {migration.id} state updated to {migration.state}")

            elif execution_state in ['ERROR', 'FAILED']:
                LOG.info(f"⚙ Migration {migration.id} failed (execution: {migration.workflow_exec}, state: {execution_state})")
                migration.state = MigrationState.ERROR
                session.commit()
                LOG.info(f"⚙ Migration {migration.id} state updated to ERROR")
            else:
                LOG.debug(f"🐛 Migration {migration.id} execution {migration.workflow_exec} still running (state: {execution_state})")

        except Exception as e:
            LOG.error(f"⧱ Error checking execution {migration.workflow_exec} for migration {migration.id}: {e}", exc_info=True)

    def schedule_new_migrations(self):
        """Schedule new migrations based on availability and timing."""
        LOG.debug(f"🐛 Scheduling new migrations")

        available_slots = self.get_available_slots()
        if available_slots <= 0:
            LOG.debug(f"🐛 No available slots for scheduling")
            return

        # Process scheduled migrations matching current hour
        remaining_slots = self.process_scheduled_migrations(available_slots)

        # Process opportunistic migrations if slots remain and within window
        if remaining_slots > 0 and self.is_within_migration_window():
            self.process_opportunistic_migrations(remaining_slots)

    def get_available_slots(self):
        """Calculate available execution slots based on actual running executions."""
        try:
            # Get currently running executions from Mistral with timeout
            running_executions = self.mistral_client.list_executions(
                state='RUNNING'
            )
            
            if running_executions is None:
                LOG.warning(f"⚠ Failed to get running executions from Mistral, using default slots")
                return 0  # Return 0 to prevent scheduling when Mistral is unreachable
                
            current_running = len(running_executions)
            available_slots = self.CONF.worker.parallel_jobs - current_running
            
            # Prevent returning negative number of job in case
            # mistral_client.list_executions() returns more than
            # CONF.worker.parallel_jobs jobs.
            available_slots = max(0, available_slots)
            
            LOG.debug(f"🐛 Currently running: {current_running}, Available slots: {available_slots}")
            
            return available_slots
            
        except Exception as e:
            LOG.error(f"⧱ Error calculating available slots: {e}")
            # Safe fallback - return 0 to prevent overloading
            return 0

    def process_scheduled_migrations(self, slots_available):
        """Process SCHEDULED migrations matching current date and hour.
        
        Uses database LIMIT to restrict the number of migrations processed
        to the number of available slots, ensuring we don't exceed capacity.
        Only processes migrations scheduled for current date and hour (ignoring minutes/seconds).
        """
        from sqlalchemy import func
        
        current_datetime = datetime.datetime.now()
        current_date = current_datetime.date()
        current_hour = current_datetime.hour
        LOG.debug(f"🐛 Processing scheduled migrations for {current_date} hour {current_hour}, slots available: {slots_available}")

        session = get_db_session()

        # Find SCHEDULED migrations for current date and hour, limited to available slots
        # Compare date and hour separately to ignore minutes/seconds
        scheduled_migrations = session.query(VMMigration).filter(
            VMMigration.state == MigrationState.SCHEDULED,
            VMMigration.scheduled_time.isnot(None),
            func.date(VMMigration.scheduled_time) == current_date,
            func.extract('hour', VMMigration.scheduled_time) == current_hour
        ).limit(slots_available).all()

        processed_count = 0
        for migration in scheduled_migrations:
            if self.trigger_migration_workflow(session, migration):
                processed_count += 1
                    
        LOG.info(f"⚙ Processed {processed_count} scheduled migrations")
        return slots_available - processed_count

    def process_opportunistic_migrations(self, slots_available):
        """Process opportunistic migrations within time window.
        
        Uses database LIMIT to restrict the number of migrations processed
        to the number of available slots, ensuring we don't exceed capacity.
        """
        LOG.debug(f"🐛 Processing opportunistic migrations, slots available: {slots_available}")

        session = get_db_session()

        # Find any SCHEDULED migrations without specific schedule time, limited to available slots
        opportunistic_migrations = session.query(VMMigration).filter(
            VMMigration.state == MigrationState.SCHEDULED,
            VMMigration.scheduled_time.is_(None)
        ).limit(slots_available).all()
        
        processed_count = 0
        for migration in opportunistic_migrations:
            if self.trigger_migration_workflow(session, migration):
                processed_count += 1

        LOG.info(f"⚙ Processed {processed_count} opportunistic migrations")

    def is_within_migration_window(self):
        """Check if current time is within allowed migration window and weekday."""
        try:
            now = datetime.datetime.now()
            
            # Parse time window
            start_time = datetime.datetime.strptime(
                self.CONF.worker.migration_start_time, '%H:%M').time()
            end_time = datetime.datetime.strptime(
                self.CONF.worker.migration_end_time, '%H:%M').time()
            
            # Handle overnight windows (e.g., 22:00 to 05:00)
            if start_time > end_time:
                # Overnight window
                current_time = now.time()
                within_time = current_time >= start_time or current_time <= end_time
                
                if within_time:
                    # Check if this overnight period crosses into weekend
                    if current_time >= start_time:
                        # We're in the evening portion - check if tomorrow is weekend
                        tomorrow = now + datetime.timedelta(days=1)
                        if now.weekday() >= 6 or tomorrow.weekday() >= 5:
                            LOG.debug(f"🐛 Skipping migration: overnight window crosses into weekend")
                            return False
                    else:
                        # We're in the morning portion - check if today is weekend
                        if now.weekday() >= 5:  # Today is Saturday or Sunday
                            LOG.debug(f"🐛 Skipping migration: morning portion of overnight window on weekend")
                            return False
            else:
                # Same-day window
                current_time = now.time()
                within_time = start_time <= current_time <= end_time
                
                if within_time:
                    # Check if today is weekend
                    if now.weekday() >= 5:  # Saturday or Sunday
                        LOG.debug(f"🐛 Skipping migration: same-day window on weekend")
                        return False
            
            if not within_time:
                LOG.debug(f"🐛 Outside migration time window ({self.CONF.worker.migration_start_time}-{self.CONF.worker.migration_end_time})")
                
            return within_time
                
        except Exception as e:
            LOG.warning(f"⚠ Error parsing migration window times: {e}")
            return False  # Safe default

    def trigger_migration_workflow(self, session, migration):
        """Trigger Mistral workflow for a migration."""
        try:
            # Prepare workflow input (only vm_id, not vm_name)
            workflow_input = {
                'vm_id': migration.vm_id
            }

            # Start workflow execution
            execution = self.mistral_client.start_workflow(workflow_input)

            if not execution:
                LOG.error(f"⧱ Failed to start workflow for migration {migration.id}")
                return False

            execution_id = execution.get('id')
            if not execution_id:
                LOG.error(f"⧱ Workflow execution missing ID for migration {migration.id}")
                return False

            # Update migration record
            migration.state = MigrationState.MIGRATING
            migration.workflow_exec = execution_id
            session.commit()

            LOG.info(f"⚙ Started migration {migration.id} with workflow execution {execution_id}")
            return True

        except Exception as e:
            LOG.error(f"⧱ Error triggering workflow for migration {migration.id}: {e}", exc_info=True)
            return False


def main():
    """Main entry point for the worker service."""
    try:
        # Initialize configuration
        CONF = config.init_config()

        # Setup logging
        logging.setup(CONF, 'vmms-worker')
        LOG.info(f"⚙ VMMS worker is starting up...")

        # Create and start worker service
        worker = WorkerService(CONF)

        # Start the worker
        success = worker.start()

        if success:
            LOG.info(f"⚙ VMMS worker stopped normally")
            sys.exit(0)
        else:
            LOG.error(f"⧱ VMMS worker failed to start")
            sys.exit(1)

    except Exception as e:
        LOG.error(f"⧱ Fatal error in worker service: {e}", exc_info=True)
        sys.exit(1)



# Copyright (c) 2025 Thomas Goirand <zigo@debian.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from sqlalchemy import Column, String, DateTime, Enum
from sqlalchemy.ext.declarative import declarative_base
from oslo_db.sqlalchemy import models
import enum
from datetime import datetime

Base = declarative_base()

class MigrationState(enum.Enum):
    SCHEDULED = "SCHEDULED"
    ERROR = "ERROR"
    MIGRATING = "MIGRATING"
    MIGRATED = "MIGRATED"

class VMMigration(Base, models.TimestampMixin):
    __tablename__ = 'vm_migrations'

    id = Column(String(36), primary_key=True)
    vm_id = Column(String(36), nullable=False)
    vm_name = Column(String(255), nullable=False)
    scheduled_time = Column(DateTime, nullable=True)
    state = Column(Enum(MigrationState), 
                   default=MigrationState.SCHEDULED,
                   nullable=False)
    workflow_exec = Column(String(36), nullable=True)

    def to_dict(self):
        return {
            'id': self.id,
            'vm_id': self.vm_id,
            'vm_name': self.vm_name,
            'scheduled_time': self.scheduled_time.isoformat() if self.scheduled_time else None,
            'state': self.state.value if isinstance(self.state, MigrationState) else self.state,
            'workflow_exec': self.workflow_exec,
            'created_at': self.created_at.isoformat() if self.created_at else None,
            'updated_at': self.updated_at.isoformat() if self.updated_at else None
        }



import unittest
from unittest import mock
import json

from flask import Flask
from oslo_context import context
from oslo_policy import policy as oslo_policy
from oslo_policy import _checks

from vmms.policy import rules
from vmms import api


class VMMSBaseTestCase(unittest.TestCase):
    """Base test case for VMMS tests."""
    pass


class VMMSTestCase(VMMSBaseTestCase):
    """Base test case for VMMS API tests."""

    def setUp(self):
        super(VMMSTestCase, self).setUp()
        
        # Create test Flask app
        self.app = api.app
        self.app.config['TESTING'] = True
        self.client = self.app.test_client()
        
        # Parse and register our rules manually for testing
        self.parsed_rules = {}
        for rule_def in rules.list_rules():
            self.parsed_rules[rule_def.name] = rule_def.check

    def _create_test_headers(self, roles=None, user_id='test_user', project_id='test_project'):
        """Create headers that simulate Keystone authentication."""
        roles = roles or []
        return {
            'X-Identity-Status': 'Confirmed',
            'X-User-Id': user_id,
            'X-Project-Id': project_id,
            'X-Roles': ','.join(roles)
        }

    def _create_unauthenticated_headers(self):
        """Create headers for unauthenticated requests."""
        return {
            'X-Identity-Status': 'Invalid'
        }


class VMMSPolicyTestCase(VMMSBaseTestCase):
    """Base test case for policy rule testing."""

    def setUp(self):
        super(VMMSPolicyTestCase, self).setUp()
        
        # Parse and register our rules manually for testing
        self.parsed_rules = {}
        for rule_def in rules.list_rules():
            self.parsed_rules[rule_def.name] = rule_def.check

    def _create_creds(self, roles=None, user_id='test_user', project_id='test_project'):
        """Create credentials dict for testing."""
        roles = roles or []
        return {
            'roles': roles,
            'user_id': user_id,
            'project_id': project_id
        }

    def _create_target(self, project_id='test_project', user_id='test_user'):
        """Create a target for policy enforcement."""
        return {
            'project_id': project_id,
            'user_id': user_id
        }

    def assertRuleIsEnforced(self, rule_name, creds, target, should_pass=True):
        """Assert that a rule is enforced correctly."""
        self.assertIn(rule_name, self.parsed_rules)
        rule = self.parsed_rules[rule_name]  # FIXED: was rule_def.name, should be rule_name
        
        # Create a minimal mock enforcer
        class MockEnforcer:
            def __init__(self, parsed_rules):
                self.rules = parsed_rules
        
        mock_enforcer = MockEnforcer(self.parsed_rules)
        
        result = _checks._check(rule, target, creds, mock_enforcer, rule_name)
        if should_pass:
            self.assertTrue(result, f"Rule {rule_name} should have passed")
        else:
            self.assertFalse(result, f"Rule {rule_name} should have failed")



# Copyright (c) 2025 Thomas Goirand <zigo@debian.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Management utility for VM Migration Scheduler."""

import sys
import os
from oslo_config import cfg
from oslo_log import log as logging

from vmms import config

def do_db_sync():
    """Sync the database up to the most recent version."""
    # Set up migration configuration
    import alembic.config
    
    # Get the absolute path to alembic.ini
    alembic_ini_path = os.path.join(
        os.path.dirname(__file__), 
        '..', 'db', 'sqlalchemy', 'alembic', 'alembic.ini'
    )
    alembic_ini_path = os.path.abspath(alembic_ini_path)
    
    # Use alembic directly
    alembic_cfg = alembic.config.Config(alembic_ini_path)
    
    # Get database connection from config
    CONF = cfg.CONF
    db_connection = CONF.database.connection
    
    # Set the database URL in alembic config
    alembic_cfg.set_main_option("sqlalchemy.url", db_connection)
    
    alembic.command.upgrade(alembic_cfg, 'head')
    
    LOG = logging.getLogger(__name__)
    LOG.info("Database sync completed successfully")

def add_command_parsers(subparsers):
    CommandParser = subparsers.add_parser
    parser = CommandParser('db_sync', help='Sync the database up to the most recent version.')
    parser.set_defaults(func=do_db_sync)

def main():
    """Parse options and call the appropriate class/method."""
    # Create CONF object first
    CONF = cfg.CONF
    
    # Register CLI options before parsing
    CONF.register_cli_opts([
        cfg.SubCommandOpt('command',
                          title='Commands',
                          help='Available commands',
                          handler=add_command_parsers)
    ])
    
    # Initialize config with the CONF object that already has CLI options
    CONF = config.init_config(conf=CONF)
    logging.setup(CONF, 'vmms-manage')
    
    # Parse command line arguments
    CONF(sys.argv[1:], project='vmms')
    
    # Execute the command
    if hasattr(CONF.command, 'func'):
        CONF.command.func()
    else:
        print("No valid command specified. Use --help for usage.")
        sys.exit(1)



# Copyright (c) 2025 Thomas Goirand <zigo@debian.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import uuid
from datetime import datetime
import ssl
import re

from flask import Flask, request, jsonify
from requests.adapters import HTTPAdapter
from sqlalchemy.exc import IntegrityError, DisconnectionError, OperationalError

from oslo_log import log as logging
from openstack import connection
from keystoneauth1 import loading as ks_loading

from vmms.db.sqlalchemy.database import init_database_config
from vmms.db.sqlalchemy.database import get_session as get_db_session
from vmms.db.sqlalchemy.models import VMMigration, MigrationState
from vmms.mistral_client import MistralClient
from vmms.policy import enforcer as policy_enforcer

# Initialize logging
LOG = logging.getLogger(__name__)

# Global config storage to avoid re-parsing
_CONFIG = None

class LegacyTLSAdapter(HTTPAdapter):
    """Custom adapter that allows legacy TLS protocols"""
    def init_poolmanager(self, *args, **kwargs):
        # Create SSL context that allows older TLS versions
        context = ssl.SSLContext(ssl.PROTOCOL_TLS)
        # Remove the SECLEVEL restriction to allow older protocols
        context.set_ciphers('DEFAULT')
        # Explicitly set minimum version to allow TLS 1.0
        context.minimum_version = ssl.TLSVersion.TLSv1
        context.maximum_version = ssl.TLSVersion.TLSv1_2
        
        kwargs['ssl_context'] = context
        return super().init_poolmanager(*args, **kwargs)

def is_valid_uuid(uuid_string):
    """Check if string is a valid UUID format"""
    try:
        uuid.UUID(uuid_string)
        return True
    except ValueError:
        return False

def is_valid_iso_datetime(datetime_string):
    """Check if string is a valid ISO datetime format"""
    try:
        datetime.fromisoformat(datetime_string.replace('Z', '+00:00'))
        return True
    except ValueError:
        return False

def validate_required_fields(data, required_fields):
    """Validate that all required fields are present"""
    return all(field in data and data[field] is not None for field in required_fields)

def get_config():
    """Get configuration, parsing it only once."""
    global _CONFIG
    if _CONFIG is None:
        from vmms import config
        _CONFIG = config.init_config()
    return _CONFIG

def get_vm_details_from_source_cloud(vm_identifier):
    # Get configuration (reuse existing config)
    CONF = get_config()

    try:
        auth         = ks_loading.load_auth_from_conf_options(CONF, "source_cloud")
        base_session = ks_loading.load_session_from_conf_options(CONF, "source_cloud", auth=auth, verify=False)
        if CONF.accept_legacy_tls:
            legacy_adapter = LegacyTLSAdapter()
            base_session.session.mount('https://', legacy_adapter)
        adapter      = ks_loading.load_adapter_from_conf_options(CONF, "source_cloud", session=base_session, auth=auth)
    except:
        LOG.error(f"Cloud not load configuration for source_cloud.")
        return None, None

    try:
        conn = connection.Connection(
            session=base_session,
            region_name=adapter.region_name,
            interface=adapter.interface,
        )
    except:
        LOG.error(f"Cloud not connect to source_cloud.")
        return None, None
            
    # Try to find VM by UUID first
    try:
        if is_valid_uuid(vm_identifier):
            servers = list(conn.compute.servers(all_projects=True, uuid=vm_identifier))
        else:
            servers = list(conn.compute.servers(all_projects=True, name=vm_identifier))
        if len(servers) == 1:
            server = servers[0]
            return server.id, server.name
        elif len(servers) > 1:
            LOG.error(f"Multiple VMs found with name: {vm_identifier} across all projects")
            return None, None
        return server.id, server.name

    except Exception as e:
        LOG.error(f"Failed to fetch VM details from source cloud: {e}", exc_info=True)
        return None, None

# Create single Flask app with all routes
app = Flask(__name__)

# wrapper decorator, so we can pass the CONF
# object ot the vmms/policy/enforcer.py
def require_vmms_policy(action):
    def decorator(f):
        @wraps(f)
        def wrapper(*args, **kwargs):
            CONF = get_config()
            return policy_enforcer.require_policy(CONF, action)(f)(*args, **kwargs)
        return wrapper
    return decorator

policy_checker = policy_enforcer.require_policy_factory(get_config)

@app.route('/healthcheck', methods=['GET'])
def health_check():
    """Health check endpoint - no authentication required"""
    return jsonify({'status': 'ok', 'service': 'vmms'}), 200

@app.route('/v2/vms', methods=['POST'])
@policy_checker('vmms:add')
def add_vm():
    """Add a VM to migration scheduler - authentication required"""
    try:
        data = request.get_json()
        
        if not data or 'vm_identifier' not in data:
            return jsonify({'error': 'Missing required field: vm_identifier'}), 400

        # Validate vm_identifier is not empty
        if not data['vm_identifier'] or not isinstance(data['vm_identifier'], str):
            return jsonify({'error': 'vm_identifier must be a non-empty string'}), 400
        
        vm_identifier = data['vm_identifier']
        LOG.info(f"Adding VM with identifier: {vm_identifier}")
        
        # Fetch VM details from source cloud
        vm_id, vm_name = get_vm_details_from_source_cloud(vm_identifier)
        
        if not vm_id or not vm_name:
            return jsonify({'error': f'Could not find VM with identifier: {vm_identifier}'}), 404
        
        LOG.info(f"Found VM: {vm_name} ({vm_id})")
        
        
        session = get_db_session()
        migration = VMMigration()
        migration.id = str(uuid.uuid4())
        migration.vm_id = vm_id
        migration.vm_name = vm_name
        
        if 'scheduled_time' in data and data['scheduled_time']:
            try:
                migration.scheduled_time = datetime.fromisoformat(data['scheduled_time'].replace('Z', '+00:00'))
            except ValueError:
                return jsonify({'error': 'Invalid scheduled_time format'}), 400
        
        migration.state = MigrationState.SCHEDULED
        
        session.add(migration)
        session.commit()
        
        result = migration.to_dict()
        return jsonify(result), 201
    except IntegrityError as e:
        session.rollback()
        if 'uniq_vm_id' in str(e) or 'vm_id' in str(e).lower():
           return jsonify({'error': f'VM with ID {vm_id} is already scheduled for migration'}), 409
        return jsonify({'error': 'Database integrity error'}), 500
    except Exception as e:
        LOG.error(f"Error adding VM: {e}", exc_info=True)
        if 'session' in locals():
            session.rollback()
        return jsonify({'error': str(e)}), 500

@app.route('/v2/vms', methods=['GET'])
@policy_checker('vmms:list')
def list_vms():
    """List all VM migrations - authentication required"""
    try:
        # Get query parameters
        state_filter = request.args.get('state', None)

        # Validate state parameter if provided
        if state_filter:
            # Get all valid state values from the enum
            valid_states = [state.value for state in MigrationState]
            if state_filter not in valid_states:
                return jsonify({
                    'error': f'Invalid state. Allowed: {", ".join(valid_states)}'
                }), 400

        session = get_db_session()

        # Build query with optional state filtering
        query = session.query(VMMigration)
        if state_filter:
            query = query.filter(VMMigration.state == state_filter)
        migrations = query.all()
        result = [migration.to_dict() for migration in migrations]
        return jsonify(result)
    except Exception as e:
        LOG.error(f"Error listing VMs: {e}", exc_info=True)
        return jsonify({'error': str(e)}), 500

@app.route('/v2/vms/<string:migration_id>', methods=['DELETE'])
@policy_checker('vmms:remove')
def remove_vm(migration_id):
    """Remove a VM from migration scheduler - authentication required"""
    # Validate migration_id format
    if not is_valid_uuid(migration_id):
        return jsonify({'error': 'Invalid migration ID format. Must be a valid UUID.'}), 400

    try:
        session = get_db_session()
        migration = session.query(VMMigration).filter_by(id=migration_id).first()
        
        if not migration:
            return jsonify({'error': f'Migration with ID {migration_id} not found'}), 404
            
        session.delete(migration)
        session.commit()
        
        return jsonify({'message': 'Migration removed successfully'}), 200
        
    except Exception as e:
        LOG.error(f"Error removing VM: {e}", exc_info=True)
        if 'session' in locals():
            session.rollback()
        return jsonify({'error': str(e)}), 500


@app.route('/v2/vms/<string:migration_id>', methods=['GET'])
@policy_checker('vmms:show')
def show_vm(migration_id):
    """Show details of a specific VM migration - authentication required"""
    # Validate migration_id format
    if not is_valid_uuid(migration_id):
        return jsonify({'error': 'Invalid migration ID format. Must be a valid UUID.'}), 400
        
    try:
        session = get_db_session()
        migration = session.query(VMMigration).filter_by(id=migration_id).first()
        
        if not migration:
            return jsonify({'error': f'Migration with ID {migration_id} not found'}), 404
            
        result = migration.to_dict()
        return jsonify(result), 200
        
    except Exception as e:
        LOG.error(f"Error showing VM: {e}", exc_info=True)
        return jsonify({'error': str(e)}), 500


@app.route('/v2/vms/<string:migration_id>', methods=['PUT'])
@policy_checker('vmms:update')
def update_vm(migration_id):
    """Update a VM migration - authentication required"""
    LOG.debug(f"🐛 API: update_vm called by user with migration_id: {migration_id}")
    LOG.debug(f"🐛 DEBUG: This should NEVER be reached if policy enforcement works!")

    # Validate migration_id format
    if not is_valid_uuid(migration_id):
        return jsonify({'error': 'Invalid migration ID format. Must be a valid UUID.'}), 400
        
    try:
        data = request.get_json()
        
        if not data:
            return jsonify({'error': 'No data provided'}), 400
            
        # Define allowed fields for update
        allowed_fields = ['scheduled_time', 'state']
        if not any(field in data for field in allowed_fields):
            return jsonify({'error': 'Missing field: scheduled_time or state'}), 400

        # Validate that only allowed fields are provided
        invalid_fields = [field for field in data.keys() if field not in allowed_fields]
        if invalid_fields:
            return jsonify({'error': f'Invalid fields: {", ".join(invalid_fields)}. Allowed: {", ".join(allowed_fields)}'}), 400

        # Validate scheduled_time format if provided
        if 'scheduled_time' in data and data['scheduled_time'] is not None:
            if not isinstance(data['scheduled_time'], str):
                return jsonify({'error': 'scheduled_time must be a string in ISO format'}), 400
            if not is_valid_iso_datetime(data['scheduled_time']):
                return jsonify({'error': 'Invalid scheduled_time format. Must be ISO 8601 format.'}), 400
        
        
        session = get_db_session()
        migration = session.query(VMMigration).filter_by(id=migration_id).first()
        
        if not migration:
            return jsonify({'error': f'Migration with ID {migration_id} not found'}), 404
            
        # Handle scheduled_time updates
        if 'scheduled_time' in data:
            # Allow unsetting scheduled_time for any state, but setting it only for scheduled migrations
            if data['scheduled_time'] is not None and migration.state != MigrationState.SCHEDULED:
                return jsonify({'error': 'Can only update scheduled_time for scheduled migrations'}), 400
                
            try:
                if data['scheduled_time']:
                    migration.scheduled_time = datetime.fromisoformat(data['scheduled_time'].replace('Z', '+00:00'))
                else:
                    migration.scheduled_time = None
            except ValueError:
                return jsonify({'error': 'Invalid scheduled_time format'}), 400
        
        # Handle state updates
        if 'state' in data:
            # Get all possible state values from the enum
            allowed_states = [state.value for state in MigrationState]
            if data['state'] not in allowed_states:
                return jsonify({'error': f'Invalid state. Allowed: {", ".join(allowed_states)}'}), 400
            migration.state = data['state']

        # Handle workflow_exec reset when state is set to SCHEDULED
        if 'state' in data and data['state'] == MigrationState.SCHEDULED.value:
            migration.workflow_exec = None

        # Validate that workflow_exec is not being set via API
        if 'workflow_exec' in data:
            return jsonify({'error': 'workflow_exec is read-only and cannot be set via API'}), 400
        
        session.commit()
        result = migration.to_dict()
        return jsonify(result), 200
        
    except Exception as e:
        LOG.error(f"Error updating VM: {e}", exc_info=True)
        if 'session' in locals():
            session.rollback()
        return jsonify({'error': str(e)}), 500


@app.route('/v2/vms/<string:migration_id>/output', methods=['GET'])
@policy_checker('vmms:output')
def get_vm_output(migration_id):
    """Get output from VM migration - authentication required"""
    LOG.debug(f"🐛 GET /v2/vms/{migration_id}/output called")

    # Validate migration_id format
    if not is_valid_uuid(migration_id):
        LOG.debug(f"🐛 Invalid UUID format: {migration_id}")
        return jsonify({'error': 'Invalid migration ID format. Must be a valid UUID.'}), 400

    try:
        # Get database session
        session = get_db_session()
        migration = session.query(VMMigration).filter_by(id=migration_id).first()

        if not migration:
            LOG.debug(f"🐛 Migration not found: {migration_id}")
            return jsonify({'error': f'Migration with ID {migration_id} not found'}), 404

        # Check if migration has a workflow execution
        if not migration.workflow_exec:
            LOG.debug(f"🐛 No workflow execution for migration: {migration_id}")
            return jsonify({'error': f'Migration {migration_id} has no workflow execution'}), 404

        # Debug state values
        LOG.debug(f"🐛 Migration state: '{migration.state}' (type: {type(migration.state)})")
        LOG.debug(f"🐛 MIGRATED: '{MigrationState.MIGRATED}' (type: {type(MigrationState.MIGRATED)})")
        LOG.debug(f"🐛 ERROR: '{MigrationState.ERROR}' (type: {type(MigrationState.ERROR)})")

        # Check if migration state allows output retrieval (must be MIGRATED or ERROR)
        allowed_states = [MigrationState.MIGRATED, MigrationState.ERROR]
        LOG.debug(f"🐛 Allowed states: {allowed_states}")
        LOG.debug(f"🐛 State check result: {migration.state in allowed_states}")

        if migration.state not in allowed_states:
            LOG.debug(f"🐛 Invalid state for output retrieval: {migration.state}")
            return jsonify({
                'error': f'Cannot retrieve output for migration in state {migration.state}. '
                         f'Output is only available for MIGRATED or ERROR states.'
            }), 400

        # Get Mistral client and retrieve output
        CONF = get_config()
        from vmms.mistral_client import MistralClient
        mistral_client = MistralClient(CONF)

        # Get output from Mistral execution
        output = mistral_client.get_migration_output(migration.workflow_exec)
        LOG.debug(f"✅ Successfully retrieved output for migration: {migration_id}")

        return jsonify(output), 200
        
    except Exception as e:
        LOG.error(f"⧱ Error getting VM output for migration {migration_id}: {e}", exc_info=True)
        return jsonify({'error': f'Failed to retrieve migration output: {str(e)}'}), 500


def get_application():
    """Create WSGI application with Keystone authentication"""
    try:
        CONF = get_config()
        logging.setup(CONF, 'vmms-api')
        
        # Initialize database configuration
        init_database_config(CONF)

        # Apply Keystone middleware with delay_auth_decision=true
        from keystonemiddleware import auth_token
        middleware_conf = dict(CONF.keystone_authtoken)
        # Ensure delay_auth_decision is true for selective auth
        middleware_conf['delay_auth_decision'] = True
        application = auth_token.filter_factory(middleware_conf)(app)
        return application
    except Exception as e:
        LOG.error(f"Error creating application: {e}", exc_info=True)
        return app



"""Initial migration

Revision ID: initial
Revises: 
Create Date: 2023-01-01 00:00:00.000000

"""
from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = 'initial'
down_revision = None
branch_labels = None
depends_on = None

def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table('vm_migrations',
        sa.Column('id', sa.String(length=36), nullable=False),
        sa.Column('vm_id', sa.String(length=36), nullable=False),
        sa.Column('vm_name', sa.String(length=255), nullable=False),
        sa.UniqueConstraint('vm_id', name='uniq_vm_id'),
        sa.Column('scheduled_time', sa.DateTime(), nullable=True),
        sa.Column('state', sa.Enum('SCHEDULED', 'MIGRATING', 'MIGRATED', 'ERROR', name='migrationstate'), nullable=True),
        sa.Column('workflow_exec', sa.String(length=36), nullable=True),
        sa.Column('created_at', sa.DateTime(), nullable=True),
        sa.Column('updated_at', sa.DateTime(), nullable=True),
        sa.PrimaryKeyConstraint('id')
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table('vm_migrations')
    # ### end Alembic commands ###



# Copyright (c) 2025 Thomas Goirand <zigo@debian.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from keystoneauth1 import loading as ks_loading
from oslo_config import cfg
from oslo_log import log as logging
from mistralclient.api import client as mistral_client

LOG = logging.getLogger(__name__)

class MistralClient(object):
    """Client for interacting with Mistral workflow service."""
    
    def __init__(self, CONF):
        self.CONF = CONF
        self.client = None
        self._initialize_client()
    
    def _initialize_client(self):
        """Initialize Mistral client."""
        try:
            # Determine which configuration section to use
            if self.CONF.worker.mistral_use_keystone_auth:
                mistralclient_auth_section = "keystone_authtoken"
                LOG.info(f"⚙ Using keystone_authtoken section for Mistral authentication")
            else:
                mistralclient_auth_section = "mistral"
                LOG.info(f"⚙ Using mistral section for Mistral authentication")
            
            LOG.debug(f"🐛 Loading auth from section: {mistralclient_auth_section}")
            
            # Load auth
            auth = ks_loading.load_auth_from_conf_options(self.CONF, mistralclient_auth_section)
            LOG.debug(f"🐛 Auth loaded successfully: {type(auth)}")
            
            # Load session with fallback for missing timeout option
            try:
                session = ks_loading.load_session_from_conf_options(
                    self.CONF, mistralclient_auth_section, auth=auth, verify=False)
            except Exception as session_e:
                LOG.debug(f"🐛 Falling back to manual session creation due to: {session_e}")
                from keystoneauth1 import session as ka_session
                session = ka_session.Session(auth=auth, verify=False)
            
            LOG.debug(f"🐛 Session loaded successfully")
            
            # Create adapter - pass both session and auth like in your working example
            try:
                adapter = ks_loading.load_adapter_from_conf_options(
                    self.CONF, mistralclient_auth_section, session=session, auth=auth)
            except Exception as adapter_e:
                LOG.debug(f"🐛 Falling back to manual adapter creation due to: {adapter_e}")
                # Get basic adapter configuration from the section
                section_conf = getattr(self.CONF, mistralclient_auth_section)
                interface = getattr(section_conf, 'interface', 'public')
                region_name = getattr(section_conf, 'region_name', None)
                
                # Manual adapter creation
                from keystoneauth1.adapter import Adapter
                adapter = Adapter(
                    session=session,
                    auth=auth,
                    service_type='workflowv2',
                    interface=interface,
                    region_name=region_name
                )
            
            LOG.debug(f"🐛 Adapter loaded successfully")
            
            # Get Mistral URL - use session catalog instead of adapter.catalog
            try:
                # Try endpoint_override first
                if hasattr(adapter, 'endpoint_override') and adapter.endpoint_override:
                    mistral_url = adapter.endpoint_override
                else:
                    # Use session to get URL from service catalog
                    mistral_url = session.get_endpoint(service_type='workflowv2', interface=adapter.interface, region_name=adapter.region_name)
            except Exception as url_e:
                LOG.error(f"⧱ Failed to discover Mistral URL: {url_e}")
                # Fallback: try to get from configuration
                section_conf = getattr(self.CONF, mistralclient_auth_section)
                mistral_url = getattr(section_conf, 'mistral_url', None)
                if not mistral_url:
                    raise Exception(f"Could not determine Mistral URL: {url_e}")
            
            LOG.debug(f"🐛 Mistral URL: {mistral_url}")
            
            # Create client
            self.client = mistral_client.client(
                mistral_url=mistral_url,
                session=session
            )
            LOG.debug(f"🐛 Mistral client created successfully")
            
            LOG.info(f"⚙ Successfully initialized Mistral client using {mistralclient_auth_section} section")
            
        except Exception as e:
            LOG.error(f"⧱ Failed to initialize Mistral client: {e}", exc_info=True)
            self.client = None
    
    def start_workflow(self, workflow_input=None):
        """Start a Mistral workflow execution using configured workflow name."""
        if not self.client:
            LOG.error(f"⧱ Mistral client not initialized")
            return None
            
        try:
            workflow_name = self.CONF.worker.mistral_workflow_name
            LOG.debug(f"🐛 Starting workflow {workflow_name}")
            LOG.debug(f"🐛 Workflow input: {workflow_input}")
            
            # Use the correct parameter names from the signature
            execution = self.client.executions.create(
                wf_identifier=workflow_name,
                workflow_input=workflow_input or {}
            )
                
            LOG.debug(f"🐛 Started workflow {workflow_name} with execution ID {execution.id}")
            return execution.to_dict()
        except Exception as e:
            LOG.error(f"⧱ Failed to start workflow {workflow_name}: {e}")
            import traceback
            LOG.error(f"⧱ Full traceback: {traceback.format_exc()}")
            return None
    
    def get_execution(self, execution_id):
        """Get workflow execution details."""
        if not self.client:
            LOG.error(f"⧱ Mistral client not initialized")
            return None
            
        try:
            execution = self.client.executions.get(execution_id)
            LOG.debug(f"🐛 Retrieved execution {execution_id} with state {execution.state}")
            return execution.to_dict()
        except Exception as e:
            LOG.error(f"⧱ Failed to get execution {execution_id}: {e}")
            return None

    def get_execution_output(self, execution_id):
        """Get workflow execution output details."""
        if not self.client:
            LOG.error(f"⧱ Mistral client not initialized")
            return None
            
        try:
            # Get full execution (fields parameter not supported in this version)
            execution = self.client.executions.get(execution_id)
            return execution.to_dict()
        except Exception as e:
            LOG.error(f"⧱ Failed to get execution output {execution_id}: {e}")
            return None
    
    def list_executions(self, **kwargs):
        """List workflow executions."""
        if not self.client:
            LOG.error(f"⧱ Mistral client not initialized")
            return []
            
        try:
            executions = self.client.executions.list(**kwargs)
            LOG.debug(f"🐛 Listed {len(executions)} executions")
            return [exec.to_dict() for exec in executions]
        except Exception as e:
            LOG.error(f"⧱ Failed to list executions: {e}")
            return []

    def get_migration_output(self, execution_id):
        """Get output from the execute_command task of a migration workflow.
        
        :param execution_id: Workflow execution ID
        :returns: Dictionary containing stdout, stderr, and exit_code
        """
        if not self.client:
            LOG.error(f"⧱ Mistral client not initialized")
            raise Exception("Mistral client not initialized")
            
        try:
            LOG.debug(f"🐛 Getting migration output for execution {execution_id}")
            
            # First, get all task executions for this workflow
            task_executions = self.client.tasks.list(workflow_execution_id=execution_id)
            LOG.debug(f"🐛 Found {len(task_executions)} task executions")
            
            # Find the execute_command task
            execute_command_task = None
            for task in task_executions:
                if task.name == 'execute_command':
                    execute_command_task = task
                    break
            
            if not execute_command_task:
                LOG.error(f"⧱ No execute_command task found in workflow execution {execution_id}")
                raise Exception(f"No execute_command task found in workflow execution {execution_id}")
            
            LOG.debug(f"🐛 Found execute_command task {execute_command_task.id}")
            
            # Get the full task object with result using tasks.get()
            full_task = self.client.tasks.get(execute_command_task.id)
            task_result = full_task.result
            LOG.debug(f"🐛 Task result retrieved: {type(task_result)}")
            
            # Your SSHAction returns properly structured result, just return it
            if isinstance(task_result, dict):
                LOG.debug(f"✅ Retrieved migration output with exit_code {task_result.get('exit_code', 0)}")
                return task_result
            else:
                # Handle string results or other formats
                LOG.warning(f"⚠ Unexpected task result type: {type(task_result)}")
                # Try to parse as JSON if it's a string
                if isinstance(task_result, str):
                    try:
                        import json
                        parsed_result = json.loads(task_result)
                        if isinstance(parsed_result, dict):
                            LOG.debug(f"✅ Parsed JSON result")
                            return parsed_result
                    except Exception:
                        LOG.debug(f"🐛 Result is string but not valid JSON")
                        pass
                # Fallback
                return {
                    'stdout': str(task_result) if task_result else '',
                    'stderr': '',
                    'exit_code': 0
                }
                
        except Exception as e:
            # Log the error for debugging
            LOG.error(f"⧱ Failed to get migration output for execution {execution_id}: {str(e)}")
            raise Exception(f"Failed to retrieve migration output: {str(e)}")



# Copyright (c) 2025 Thomas Goirand <zigo@debian.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oslo_policy import policy

# Define your custom rules with descriptions
rules = [
    # Default deny rule
    policy.RuleDefault(
        name='default',
        check_str='!',
        description='Default rule that denies all actions unless explicitly allowed'
    ),

    # Role hierarchy helper rules
    policy.RuleDefault(
        name='is_admin',
        check_str='role:admin',
        description='Check if user has admin role'
    ),
    policy.RuleDefault(
        name='is_vmms_operator',
        check_str='role:vmms_operator or role:admin',
        description='Check if user has vmms_operator or admin role (can perform all operations)'
    ),
    policy.RuleDefault(
        name='is_vmms_viewer',
        check_str='role:vmms_viewer or role:vmms_operator or role:admin',
        description='Check if user has vmms_viewer, vmms_operator, or admin role (can view operations)'
    ),
    
    # API-specific rules using helper rules (cleaner and more maintainable)
    policy.DocumentedRuleDefault(
        name='vmms:add',
        check_str='rule:is_vmms_operator',
        description='Add a VM to migration scheduler',
        operations=[
            {
                'method': 'POST',
                'path': '/v2/vms'
            }
        ]
    ),
    policy.DocumentedRuleDefault(
        name='vmms:remove',
        check_str='rule:is_vmms_operator',
        description='Remove a VM from migration scheduler',
        operations=[
            {
                'method': 'DELETE',
                'path': '/v2/vms/{migration_id}'
            }
        ]
    ),
    policy.DocumentedRuleDefault(
        name='vmms:update',
        check_str='rule:is_vmms_operator',
        description='Update VM migration settings',
        operations=[
            {
                'method': 'PUT',
                'path': '/v2/vms/{migration_id}'
            }
        ]
    ),
    policy.DocumentedRuleDefault(
        name='vmms:list',
        check_str='rule:is_vmms_viewer',
        description='List all VM migrations',
        operations=[
            {
                'method': 'GET',
                'path': '/v2/vms'
            }
        ]
    ),
    policy.DocumentedRuleDefault(
        name='vmms:show',
        check_str='rule:is_vmms_viewer',
        description='Show details of a specific VM migration',
        operations=[
            {
                'method': 'GET',
                'path': '/v2/vms/{migration_id}'
            }
        ]
    ),
    policy.DocumentedRuleDefault(
        name='vmms:output',
        check_str='rule:is_vmms_viewer',
        description='Get output from VM migration',
        operations=[
            {
                'method': 'GET',
                'path': '/v2/vms/{migration_id}/output'
            }
        ]
    ),
]

def list_rules():
    return rules



# Copyright (c) 2025 Thomas Goirand <zigo@debian.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import configparser
import os
import subprocess
import sys
import tempfile
from jinja2 import Environment, FileSystemLoader

def load_config(config_file):
    """Load configuration from file."""
    if not os.path.exists(config_file):
        raise FileNotFoundError(f"Configuration file not found: {config_file}")
    
    if not os.access(config_file, os.R_OK):
        raise PermissionError(f"Cannot read configuration file: {config_file}")
    
    config = {}
    
    # Simple key=value parser for shell-style config
    with open(config_file, 'r') as f:
        for line in f:
            line = line.strip()
            if line and not line.startswith('#'):
                if '=' in line:
                    key, value = line.split('=', 1)
                    key = key.strip()
                    value = value.strip().strip('"\'')
                    
                    # Handle command substitution like $(cat file)
                    if value.startswith('$(') and value.endswith(')'):
                        command = value[2:-1]
                        try:
                            result = subprocess.check_output(command, shell=True, text=True)
                            value = result.strip()
                        except subprocess.CalledProcessError as e:
                            raise RuntimeError(f"Failed to execute command '{command}': {e}")
                    
                    config[key] = value
    
    return config

def render_workflow_template(template_path, config):
    """Render workflow template with configuration."""
    if not os.path.exists(template_path):
        raise FileNotFoundError(f"Template file not found: {template_path}")
    
    # Read template content
    with open(template_path, 'r') as f:
        template_content = f.read()
    
    # Render with Jinja2
    env = Environment()
    template = env.from_string(template_content)
    
    # Process private key to add proper indentation for YAML
    if 'ssh_private_key_path' in config:
        try:
            with open(config['ssh_private_key_path'], 'r') as f:
                private_key_content = f.read().rstrip()
            # Add indentation for YAML multiline string
            config['ssh_private_key'] = '\n'.join([' ' * 10 + line for line in private_key_content.split('\n')])
        except Exception as e:
            raise RuntimeError(f"Failed to read SSH private key: {e}")
    
    return template.render(**config)

def upload_workflow(workflow_content):
    """Upload workflow to Mistral using openstack CLI."""
    try:
        # Create temporary file
        with tempfile.NamedTemporaryFile(mode='w', suffix='.yaml', delete=False) as f:
            f.write(workflow_content)
            temp_file = f.name
        
        try:
            # Upload using openstack CLI
            result = subprocess.run(
                ['openstack', 'workflow', 'create', temp_file],
                capture_output=True,
                text=True,
                check=True
            )
            print("Workflow uploaded successfully!")
            if result.stdout:
                print(result.stdout)
        finally:
            # Clean up temporary file
            os.unlink(temp_file)
            
    except subprocess.CalledProcessError as e:
        raise RuntimeError(f"Failed to upload workflow: {e.stderr}")
    except Exception as e:
        raise RuntimeError(f"Error during workflow upload: {e}")

def main():
    """Main entry point for workflow uploader."""
    config_file = '/etc/vmms/vmms-upload-workflow.conf'
    template_path = '/usr/share/vmms-common/workflows/migrate_vm_workflow.yaml.j2'
    
    try:
        # Load configuration
        print("Loading configuration...")
        config = load_config(config_file)
        
        # Render template
        print("Rendering workflow template...")
        workflow_content = render_workflow_template(template_path, config)
        
        # Upload workflow
        print("Uploading workflow to Mistral...")
        upload_workflow(workflow_content)
        
        print("Workflow upload completed successfully!")
        return 0
        
    except Exception as e:
        print(f"Error: {e}", file=sys.stderr)
        return 1



# Copyright (c) 2025 Thomas Goirand <zigo@debian.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oslo_policy import policy as oslo_policy
from oslo_context import context as oslo_context
from flask import request, jsonify
from functools import wraps
import logging

# Import your policies explicitly
from vmms.policy import rules as vmms_rules

LOG = logging.getLogger(__name__)

_ENFORCER = None

def init_enforcer(CONF):
    global _ENFORCER
    if not _ENFORCER:
        LOG.debug("🐛 Initializing policy enforcer")
        _ENFORCER = oslo_policy.Enforcer(CONF)

        # Explicitly register your policies
        LOG.debug("🐛 Registering VMMS policy rules")
        for rule in vmms_rules.list_rules():
            _ENFORCER.register_default(rule)

        # Force load the rules
        _ENFORCER.load_rules()

        LOG.debug(f"🐛 Registered {len(_ENFORCER.registered_rules)} policy rules")
            
    return _ENFORCER

def get_enforcer(CONF):
    enforcer = _ENFORCER or init_enforcer(CONF)
    return enforcer

def enforce_policy(CONF, action, target=None):
    """Enforce policy for the current Flask request"""
    # Create context from request headers (set by keystonemiddleware)
    roles_header = request.headers.get('X-Roles', '')
    roles = roles_header.split(',') if roles_header else []
    roles = [role.strip() for role in roles]  # Clean up whitespace
    
    ctx = oslo_context.RequestContext(
        user_id=request.headers.get('X-User-Id'),
        project_id=request.headers.get('X-Project-Id'),
        roles=roles,
        is_admin='admin' in roles
    )
    
    # Get policy enforcer
    enforcer = get_enforcer(CONF)
    
    # Set default target if not provided
    if target is None:
        target = {
            'project_id': ctx.project_id or '',
            'user_id': ctx.user_id or ''
        }
    
    # Get credentials from context
    creds = ctx.to_policy_values()

    try:
        # Use the standard enforce method
        result = enforcer.enforce(action, target, creds)
        
        if not result:
            raise oslo_policy.PolicyNotAuthorized(action, target, creds)
            
        return result
    except oslo_policy.PolicyNotAuthorized:
        raise
    except Exception as e:
        LOG.error(f"⧱ Policy enforcement failed for {action}: {e}", exc_info=True)
        raise

def require_policy_factory(get_config_func):
    """Factory function to create policy decorator with config getter"""
    def require_policy(action):
        """Decorator to enforce policy on API endpoints"""
        def decorator(f):
            @wraps(f)
            def wrapper(*args, **kwargs):
                # Check authentication first
                identity_status = request.headers.get('X-Identity-Status')
                if not identity_status or identity_status.upper() != 'CONFIRMED':
                    return jsonify({'error': 'Authentication required'}), 401
                
                try:
                    CONF = get_config_func()
                    enforce_policy(CONF, action)
                except KeyError as e:
                    LOG.error(f"⧱ Policy key error: {e}")
                    return jsonify({'error': 'Policy configuration error'}), 500
                except Exception as e:
                    return jsonify({'error': 'Forbidden'}), 403
                    
                return f(*args, **kwargs)
            return wrapper
        return decorator
    return require_policy



# Copyright (c) 2025 Thomas Goirand <zigo@debian.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import socket
import sys

from keystonemiddleware import auth_token
from keystoneauth1 import loading as ks_loading
from oslo_config import cfg
from oslo_db import options as db_options
from oslo_log import log as logging

vmms_default_opts = [
    cfg.BoolOpt('accept_legacy_tls',
               default=False,
               help='Accept legacy TLS connecition (ie: TLS 1.0) when '
                    'connecting to source_cloud. This is useful for '
                    'connecting to old deployments.'),
]

# Worker configuration
worker_group = cfg.OptGroup(name='worker',
                            title='Worker options')

worker_opts = [
    cfg.StrOpt('coordination_url',
               default=None,
               help='Tooz coordination URL (e.g., zookeeper://host:port)'),
    cfg.BoolOpt('mistral_use_keystone_auth',
               default=True,
               help='Use keystone_authtoken section for Mistral authentication '
                    'instead of separate mistral section.'),
    cfg.StrOpt('my_hostname',
               default=None,
               sample_default='cl1-controller-1.example.com',
               help='FQDN of this worker instance. Used for coordination. '
                    'Must be unique across the worker cluster. If None, the server FQDN is used.'),
    cfg.IntOpt('parallel_jobs',
               default=3,
               help='Maximum number of parallel Mistral workflow executions'),
    cfg.IntOpt('check_interval',
               default=120,
               help='Periodic check interval in seconds'),
    cfg.StrOpt('migration_start_time',
               default='02:00',
               help='Daily start time for opportunistic migrations (HH:MM)'),
    cfg.StrOpt('migration_end_time',
               default='06:00',
               help='Daily end time for opportunistic migrations (HH:MM)'),
    cfg.StrOpt('mistral_workflow_name',
               default='migrate_vm',
               help='Name of the Mistral workflow to execute for migrations'),
]

def init_config(conf=None):
    """Initialize configuration and return CONF object"""
    
    # Use provided CONF or create new one
    if conf is not None:
        CONF = conf
    else:
        CONF = cfg.CONF
    
    # Register migration timing options
    CONF.register_group(worker_group)
    CONF.register_opts(worker_opts, group=worker_group)

    # DEFAULT group
    CONF.register_opts(vmms_default_opts)
    
    # Register oslo.log options
    logging.register_options(CONF)
    
    # Register oslo.db options (this includes [database]/connection)
    db_options.set_defaults(CONF)
    
    # Register keystone middleware options for API auth
    auth_token.CONF = CONF
    try:
        auth_token_opts = auth_token._opts
        CONF.register_opts(auth_token_opts, group='keystone_authtoken')
    except (AttributeError, TypeError):
        # Fallback if _opts doesn't work as expected
        pass
    
    # Register keystoneauth options for source cloud connection
    ks_loading.register_session_conf_options(CONF, "source_cloud")
    ks_loading.register_auth_conf_options(CONF, "source_cloud")
    ks_loading.register_adapter_conf_options(CONF, "source_cloud")

    # Register keystoneauth options for mistral connection
    ks_loading.register_session_conf_options(CONF, "mistral")
    ks_loading.register_auth_conf_options(CONF, "mistral")
    ks_loading.register_adapter_conf_options(CONF, "mistral")

    # Load configuration from file (only if no conf was provided)
    if conf is None:
        CONF(sys.argv[1:], project='vmms')

    # Set hostname if not provided
    if CONF.worker.my_hostname is None:
        CONF.worker.my_hostname = socket.getfqdn()

    # Set up logging
    logging.setup(CONF, 'vmms')
    
    return CONF

def list_opts():
    """Return a list of oslo.config options available in this module."""
    # Get keystoneauth options properly
    auth_opts = ks_loading.get_auth_common_conf_options()
    session_opts = ks_loading.get_session_conf_options()
    adapter_opts = ks_loading.get_adapter_conf_options()

    # Get auth plugin specific options (used by both source_cloud and mistral)
    auth_plugin_opts = []
    for name in ks_loading.get_available_plugin_names():
        auth_plugin_opts.extend(ks_loading.get_auth_plugin_conf_options(name))
    
    return [
        ("DEFAULT", vmms_default_opts),
        ('worker', worker_opts),
        ('source_cloud', list(auth_opts) + auth_plugin_opts + list(session_opts) + list(adapter_opts)),
        ('mistral', list(auth_opts) + auth_plugin_opts + list(session_opts) + list(adapter_opts)),
    ]
